from asyncio import exceptions
from bergen.schema import DataPoint
from bergen.wards.graphql.aiohttp import AIOHttpGraphQLWard
from bergen.auths.base import BaseAuthBackend
from typing import Callable, Dict
from bergen.enums import DataPointType
from bergen.clients.base import BaseWard
import logging
import asyncio
from bergen.console import console
logger = logging.getLogger(__name__)

datapointregistry = None

class DataPointRegistry(object):


    def __init__(self) -> None:
        self.pointIDPointMap: Dict[str, DataPoint] = {}
        self.pointIDWardMap: Dict[str, BaseWard] = {}
        self.pointIDNegotiationMap: Dict[str, BaseWard] = {}
        self.builders =  {
                # Default Builders for standard
                DataPointType.GRAPHQL: AIOHttpGraphQLWard
        }


    def register_client_builder(self, type:str , builder: Callable):
        self.builders[type] = builder

    def create_ward(self, bergen, point) -> BaseWard:

        if point.app.name == bergen.application.name:
            console.log("[red] Datapoint App is the Same as this One, will not create own Ward!!!")
            return None

        if point.id in self.pointIDWardMap:
            return self.pointIDWardMap[point.id]

        logger.info(f"Creating new Ward for Datapoint {point}")

        if point.type in self.builders:
            builder = self.builders[point.type]
            self.pointIDWardMap[point.id]  = builder(bergen, point)
            self.pointIDPointMap[point.id]  = point

            return self.pointIDWardMap[point.id]
        else:
            raise NotImplementedError(f"We have no idea how to build the ward for this Datapoint {point.type}")

    async def configure_wards(self):
        wards = [ward for id, ward in self.pointIDWardMap.items()]
        names = [point.app.name for id, point in self.pointIDPointMap.items()]

        negotiation_results = await asyncio.gather(*[ward.configure() for ward in wards], return_exceptions=True)
        print(negotiation_results)
        extensions =  {name: result for name, result in zip(names, negotiation_results)}
        return extensions




def get_datapoint_registry() -> DataPointRegistry:
    global datapointregistry
    if datapointregistry is None:
        datapointregistry = DataPointRegistry()
    return datapointregistry




from abc import ABC, abstractmethod
from bergen.clients.base import BaseBergen

from websockets.exceptions import ConnectionClosedError
from bergen.messages import *
from bergen.messages.utils import expandToMessage
from bergen.messages.base import MessageModel
from bergen.provider.base import BaseProvider
import logging
import asyncio
import websockets
import json
import asyncio
from bergen.console import console
try:
    from asyncio import create_task
except ImportError:
    #python 3.6 fix
    create_task = asyncio.ensure_future


logger = logging.getLogger()



DEBUG = True

class WebsocketProvider(BaseProvider):
    ''' Is a mixin for Our Bergen '''

    def __init__(self, client: BaseBergen, host=None, port=None,  protocol=None, auth=None, auto_reconnect=True,**kwargs) -> None:
        super().__init__(client, **kwargs)
        self.websocket_host = host or client.config.host
        self.websocket_port = port or client.config.port
        self.websocket_protocol = protocol
        self.token = auth["token"]

        self.pending = None

        self.auto_reconnect = True
        self.allowed_retries = 2
        self.current_retries = 0


    async def connect(self):
        self.incoming_queue = asyncio.Queue()
        self.outgoing_queue = asyncio.Queue()


        self.tasks = {}

        self.startup_task = create_task(self.startup())


    async def disconnect(self) -> str:

        for task in self.pending:
            task.cancel()


        if self.connection: await self.connection.close()

        if self.pending:
            for task in self.pending:
                task.cancel()

            try:
                await asyncio.wait(self.pending)
            except asyncio.CancelledError:
                pass

            logger.info("Peasent disconnected")

        
    async def startup(self):
        try:
            await self.connect_websocket()
        except Exception as e:

            console.print("[blue] Connection attempt as Provider failed")
            self.current_retries += 1
            if self.auto_reconnect:
                sleeping_time = (self.current_retries + 1)
                console.print(f"[blue] Trying to Reconnect as Provider in {sleeping_time} seconds")
                await asyncio.sleep(sleeping_time)
                await self.startup()
            else:
                console.print("[blue]Provider: No reconnecting attempt envisioned. Shutting Down!")
                console.print_exception()

        console.print("[blue] Sucessfully Established A Connection to Provider Endpoint")

        self.consumer_task = create_task(
            self.consumer()
        )

        self.producer_task = create_task(
            self.producer()
        )

        self.worker_task = create_task(
            self.workers()
        )

        done, self.pending = await asyncio.wait(
            [self.consumer_task, self.worker_task, self.producer_task],
            return_when=asyncio.FIRST_EXCEPTION
        )

        try:
            for task in done:
                if task.exception():
                    raise task.exception()
        except ConnectionClosedError:
            console.print("[blue] Provider Connection was closed. Trying Reconnect")
        except:
            console.print_exception()
                    
        logger.error(f"Provider: Lost connection inbetween everything :( {[ task.exception() for task in done]}")
        logger.error(f'Provider: Reconnecting')

        

        if self.connection: await self.connection.close()

        for task in self.pending:
            task.cancel()

        console.log("Trying to Reconnect")
        self.current_retries = 0 # reset retries after one successfull connection
        await self.startup() # Attempt to ronnect again
        

    async def connect_websocket(self):
        uri = f"{self.websocket_protocol}://{self.websocket_host}:{self.websocket_port}/provider/?token={self.token}"
        self.connection = await websockets.client.connect(uri)
        logger.info("Sueccess fully connected Provider ")


    async def consumer(self):
        async for message in self.connection:
            await self.incoming_queue.put(message)

    async def producer(self):
        while True:
            message = await self.outgoing_queue.get()
            await self.connection.send(message.to_channels())

            self.outgoing_queue.task_done()

    async def forward(self, message: MessageModel) -> str:
        logger.info(f"Sending {message}")
        await self.outgoing_queue.put(message)

    async def workers(self):
        while True:
            message_str = await self.incoming_queue.get()
            try:
                message = expandToMessage(json.loads(message_str))
                await self.on_message(message)
            except:
                raise


            self.incoming_queue.task_done()
    




    




from bergen.console import console
from bergen.schema import Node
from bergen.graphical import GraphicalBackend


class Interaction:
    def __init__(self, node: Node) -> None:
        self.node = node

    async def graphical_assign(self):
        from bergen.ui.assignation import AssignationUI

        with console.status("[bold green]Using Graphical Assignment"):
            form = AssignationUI()
            nana = form.exec_()
        return nana


    async def __aenter__(self):
        console.log("Interaction started")
        self.graphical_backend = await GraphicalBackend().__aenter__()
        return self


    async def __aexit__(self, *args, **kwargs):
        console.log("Interaction Done")



from bergen.models import Node
from bergen.schema import NodeType


from typing import Callable, Type
from bergen.types.node.ports.arg import IntArgPort, ModelArgPort, StringArgPort
from bergen.types.node.ports.kwarg import IntKwargPort, ModelKwargPort, StringKwargPort
from bergen.types.node.ports.returns import IntReturnPort, ModelReturnPort
from bergen.types.model import ArnheimModel
import inspect
import logging
from inspect import signature, Parameter
from docstring_parser import parse

logger = logging.getLogger(__name__)



def createNodeFromActor(actor,*args, **kwargs):
    raise NotImplementedError("No longer supported")
    #return createNodeFromFunction(actor.assign, *args, interface=actor.__name__.lower())

def parseFunctionToDict(function: Callable, widgets: dict = {}, allow_empty_doc=False, interface=None):
    is_generator = inspect.isasyncgenfunction(function) or inspect.isgenerator(function)
    logger.info(f"Node is {'Generator' if is_generator else 'Function'}")

    sig = signature(function)
    # Generate Args and Kwargs from the Annotation
    args = []
    kwargs = []
    function_inputs = sig.parameters
    for key, value in function_inputs.items():
        widget = widgets.get(key, None)
        the_class = value.annotation

        if value.default == Parameter.empty: # Non default parameters are args
            if issubclass(the_class, ArnheimModel):
                args.append(ModelArgPort.fromParameter(value, the_class.getMeta(), widget=widget))
            if issubclass(the_class, int):
                args.append(IntArgPort.fromParameter(value, widget=widget))
            if issubclass(the_class, str):
                args.append(StringArgPort.fromParameter(value, widget=widget))
        else:
            if issubclass(the_class, ArnheimModel):
                kwargs.append(ModelKwargPort.fromParameter(value, the_class.getMeta(), widget=widget))
            if issubclass(the_class, int):
                kwargs.append(IntKwargPort.fromParameter(value, widget=widget))
            if issubclass(the_class, str):
                kwargs.append(StringKwargPort.fromParameter(value, widget=widget))



    # Generate types from the Output
    returns = []

    function_output = sig.return_annotation
    try:
        # Raises type error if we use it with a class but needed here because typing is actually not a class but an Generic Alias :rolling_eyes::
        if function_output._name == "Tuple":
            for type in function_output.__args__:
                if issubclass(type, ArnheimModel):
                    returns.append(ModelReturnPort(value, key=type.__name__))
                if issubclass(type, int):
                    returns.append(IntReturnPort(key=type.__name__))

    except AttributeError:
        # Once here we should have only classes... lets see about that
        if issubclass(function_output, ArnheimModel):
                returns.append(ModelReturnPort(function_output.getMeta(), key=function_output.__name__))
        if issubclass(function_output, int):
                returns.append(IntReturnPort(key=function_output.__name__))

    # Docstring Parser to help with descriptions
    docstring = parse(function.__doc__)
    if docstring.long_description is None:
        assert allow_empty_doc is not False, f"We don't allow empty documentation for function {function.__name__}. Please Provide"
        logger.warn(f"Allowing empty Documentatoin. Please consider providing a documentation for function {function.__name__}")


    name = interface or docstring.short_description or function.__name__
    description = docstring.long_description or "No Description"

    doc_param_map = {param.arg_name: {
        "required": not param.is_optional,
        "description": param.description if not param.description.startswith("[") else None,
    } for param in docstring.params}


    if docstring.returns:
        return_description = docstring.returns.description
        seperated_list = return_description.split(",")
        assert len(returns) == len(seperated_list), "Length of Description and Returns not Equal: If you provide a Return Annotation make sure you seperate the description for each port with ','"
        for index, doc in enumerate(seperated_list):
            returns[index].description = doc

    # TODO: Update with documentatoin.... (Set description for portexample)
    for port in args:
        if port.key in doc_param_map:
            updates = doc_param_map[port.key]
            port.description = updates["description"] or port.description

    for port in kwargs:
        if port.key in doc_param_map:
            updates = doc_param_map[port.key]
            port.description = updates["description"] or port.description


    logger.info(f"Creating Arg Ports: {[str(port.key) for  port in args]}")
    logger.info(f"Creating Kwargs Ports: {[str(port.key) for  port in kwargs]}")
    logger.info(f"Creating Returns Ports: {[str(port.key) for  port in returns]}")

    return {
        "name": name,
        "interface": interface or function.__name__,
        "description" : description,
        "args" : [arg.serialize() for arg in args],
        "kwargs" : [kwarg.serialize() for kwarg in kwargs],
        "returns" : [re.serialize() for re in returns],
        "type" : NodeType.GENERATOR if is_generator else NodeType.FUNCTION
    }


def createNodeFromFunction(function: Callable, widgets: dict = {}, allow_empty_doc=False, interface=None):
    node_dict = parseFunctionToDict(function, widgets=widgets, allow_empty_doc=allow_empty_doc, interface=interface)
    return Node.objects.update_or_create(**node_dict)


from bergen.enums import HostProtocol, PostmanProtocol, ProviderProtocol
from bergen.queries.delayed.pod import POD_QUERY
from bergen.queries.delayed.template import TEMPLATE_GET_QUERY
from bergen.extenders.port import ArgPortExtender, KwargPortExtender, ReturnPortExtender
from bergen.types.object import ArnheimObject
from bergen.types.model import ArnheimModel
from enum import Enum
from typing import  Any, Generic, List, Optional, Type, TypeVar
try:
	# python 3.8
	from typing import ForwardRef
except ImportError:
	# ForwardRef is private in python 3.6 and 3.7
	from typing import _ForwardRef as ForwardRef

DataModel = ForwardRef("DataModel")



class AssignationParams(ArnheimObject):
    provider: Optional[str]

class ProvisionParams(ArnheimObject):
    provider: Optional[str]
    auto_unprovide: Optional[bool]


class HerreUser(ArnheimModel):
    id: Optional[int]
    username: Optional[str]
    firstName: Optional[str]
    lastName: Optional[str]

    class Meta:
        identifier = "user"

class HerreApp(ArnheimModel):
    client_id: Optional[str]
    name: Optional[str]
    grant_type: Optional[str]

    class Meta:
        identifier = "app"

class DataPoint(ArnheimModel):
    app: Optional[HerreApp]
    user: Optional[HerreUser]
    type: Optional[str]
    name: Optional[str]
    needsNegotiation: Optional[bool]
    secure: Optional[bool]
    inward: Optional[str]
    outward: Optional[str]
    port: Optional[int]
    url: Optional[str]
    models: Optional[List[DataModel]]

    class Meta:
        identifier = "datapoint"


class DataModel(ArnheimModel):
    identifier: Optional[str]
    extenders: Optional[List[str]]
    point: Optional[DataPoint]

    class Meta:
        identifier = "datamodel"



class PostmanSettings(ArnheimObject):
    type: Optional[PostmanProtocol]
    kwargs: Optional[dict]

class ProviderSettings(ArnheimObject):
    type: Optional[ProviderProtocol]
    kwargs: Optional[dict]

class EntertainerSettings(ArnheimObject):
    type: Optional[HostProtocol]
    kwargs: Optional[dict]

class Transcript(ArnheimObject):
    postman: Optional[PostmanSettings]
    host: Optional[EntertainerSettings]
    provider: Optional[ProviderSettings]
    models: Optional[List[DataModel]]


class Provider(ArnheimModel):
    name: Optional[str]

    class Meta:
        identifier = "provider"


class Widget(ArnheimObject):
    query: Optional[str]
    dependencies: Optional[List[str]]


class ArgPort(ArgPortExtender, ArnheimObject):
    __slots__ = ("_widget", )

    required: Optional[bool]
    key: Optional[str]
    identifier: Optional[str] 
    label: Optional[str]
    widget: Optional[Widget]

class KwargPort(KwargPortExtender, ArnheimObject):
    __slots__ = ("_widget", )

    required: Optional[bool]
    key: Optional[str]
    identifier: Optional[str] 
    label: Optional[str]
    default: Optional[Any]
    widget: Optional[Widget]

class ReturnPort(ReturnPortExtender, ArnheimObject):
    __slots__ = ("_widget", )

    required: Optional[bool]
    key: Optional[str]
    identifier: Optional[str] 
    label: Optional[str]


class NodeType(str, Enum):
    FUNCTION = "FUNCTION"
    GENERATOR = "GENERATOR"


class Node(ArnheimModel):
    id: Optional[int]
    name: Optional[str]
    description: Optional[str]
    package: Optional[str]
    interface: Optional[str]
    args: Optional[List[ArgPort]]
    kwargs: Optional[List[KwargPort]]
    returns: Optional[List[ReturnPort]]
    type: Optional[NodeType]


    class Meta:
        identifier = "node"


class ArnheimApplication(ArnheimModel):
    logo: Optional[str]

    class Meta:
        identifier = "arnheim_application"


class Peasent(ArnheimModel):
    name: Optional[str]
    application: Optional[ArnheimApplication]

    class Meta:
        identifier = "peasent"


class Volunteer(ArnheimModel):
    identifier: str
    node: Node

    class Meta:
        identifier = "volunteer"

class Template(ArnheimModel):
    node: Optional[Node]
    provider: Optional[Provider]
    channel: Optional[str]

    class Meta:
        identifier = "template"
        get = TEMPLATE_GET_QUERY



class Pod(ArnheimModel):
    name: Optional[str]
    channel: Optional[str]
    template: Optional[Template]
    status: Optional[str]

    class Meta:
        identifier = "pod"

class PeasentPod(Pod):
    peasent: Optional[Peasent]

    class Meta:
        identifier = "peasentpod"


class Provision(ArnheimModel):
    pod: Optional[Pod]
    node: Optional[Node]
    status: Optional[str]
    statusmessage: Optional[str]
    reference: Optional[str]

    class Meta:
        identifier = "provision"


class AssignationStatus(str, Enum):
    ERROR = "ERROR"
    PROGRESS = "PROGRESS"
    DEBUG = "DEBUG"
    DONE = "DONE"
    YIELD = "YIELD"
    CRITICAL ="CRITICAL"
    PENDING = "PENDING"


class ProvisionStatus(str, Enum):
    ERROR = "ERROR"
    PROGRESS = "PROGRESS"
    DONE = "DONE"
    CRITICAL ="CRITICAL"
    PENDING = "PENDING"

class PodStatus(str, Enum):
    PENDING = "PENDING"
    ACTIVE = "ACTIVE"
    ERROR = "ERROR"

class Assignation(ArnheimModel):
    pod: Optional[Pod]
    id: Optional[int]
    inputs: Optional[dict]
    outputs: Optional[dict]
    message: Optional[str]
    status: Optional[str]
    statusmessage: Optional[str]

    class Meta:
        identifier = "assignation"


class VartPod(Pod):
    volunteer:  Optional[Volunteer] 

    class Meta:
        identifier = "vartpod"




Avatar.update_forward_refs()
DataPoint.update_forward_refs()
Node.update_forward_refs()



from PyQt5.QtCore import Qt
from bergen.query import ListQuery, QueryList, TypedGQL
from typing import Any, Optional
from bergen.types.object import ArnheimObject
from bergen.schema import Port
from bergen.ui.widgets.base import BaseWidgetMixin
from PyQt5.QtWidgets import QComboBox, QHBoxLayout, QLabel, QSlider, QWidget
from bergen.registries.arnheim import get_current_arnheim

class SelectorItem(ArnheimObject):
    value: Optional[Any]
    label: Optional[str]




class SliderWidget(BaseWidgetMixin, QWidget):
    query = None


    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.sld = QSlider(Qt.Horizontal, self)
        self.label = QLabel(str(self.port.default))
        self.sld.setRange(0, 100)
        self.sld.setFocusPolicy(Qt.NoFocus)
        self.sld.setPageStep(5)
        self.sld.setValue(self.port.default)
        self.sld.valueChanged.connect(self.valueChanged)

        self.layout = QHBoxLayout()
        self.layout.addWidget(self.sld)
        self.layout.addWidget(self.label)
        self.current_value = None
        self.setLayout(self.layout)


    def valueChanged(self, value):
        if value == self.current_value:
            return

        self.current_value = value
        self.label.setText(str(self.current_value))
        if self.on_changed: self.on_changed(value)


    def render(self, keyValueMap):
        pass
        
     
    @classmethod
    def fromPort(cls, port: Port, **kwargs):
        return cls(port=port, **kwargs)





import inspect
from typing import DefaultDict
from bergen.types.node.widgets.base import BaseWidget


class BaseKwargPort:

  def __init__(self, widget, key=None, label=None, description=None, default=None) -> None:
      assert isinstance(widget, BaseWidget) or widget is None, "Widget be instance of a subclass of BaseWidget or None"
      self.type = self.__class__.__name__
      self.key = key
      self.label = label
      self.widget = widget
      self.default = default
      self.description = description
      super().__init__()


  def serialize(self):
      assert self.key is not None, "Please provide at least a key to your Port"
      widgetFragment = {"widget": self.widget.serialize()} if self.widget is not None else {"widget": None}
      return {
        "type": self.type,
        "key": self.key,
        "label" : self.label or self.key.capitalize(),
        "description": self.description,
        "default": self.default,
        **widgetFragment
      }


  def __call__(self, key):
      self.key = key
      return self

  @classmethod
  def fromParameter(cls, param: inspect.Parameter, *args,  label=None, description=None, **kwargs):
      return cls(
        *args,
        default = param.default if not param.default == inspect.Parameter.empty else None,
        label = label or param.name.capitalize(),
        key = param.name,
        **kwargs
      )


  def __str__(self) -> str:
      return str(self.serialize())


from pydantic.main import BaseModel
from bergen.auths.legacy.app import LegacyApplication
from bergen.auths.backend.app import ArnheimBackendOauth
from bergen.auths.implicit.app import ImplicitApplication
from bergen.clients.base import BaseBergen
from bergen.config.types import GrantType, HerreConfig, ArkitektConfig
import os
import yaml
import logging


logger = logging.getLogger(__name__)




class Bergen(BaseBergen):

    def __init__(self,
    config_path = "bergen.yaml",
    force_new_token = False,
    arkitekt_host: str = None, 
    arkitekt_port: int = None,
    herre_host: str = None,
    herre_port: int = None,
    client_id: str = None, 
    client_secret: str = None,
    username: str = None,
    password: str = None,
    grant_type: GrantType = None,
    bind=True,
    allow_insecure=False,
    **kwargs) -> None:


        arkitekt_config = {}
        herre_config = {}

        if os.path.isfile(config_path):
            with open(config_path,"r") as file:
                config = yaml.load(file, Loader=yaml.FullLoader)

                if "arkitekt" in config:
                    arkitekt_config.update(config["arkitekt"])

                if "herre" in config:
                    herre_config.update(config["herre"])

        else:
            raise Exception("No configuration file found! (please add to the directory or specify config_path")


        if arkitekt_host : arkitekt_config["port"]= arkitekt_host
        if arkitekt_port : arkitekt_config["host"]= arkitekt_port
        if herre_host : herre_config["host"]= herre_host
        if herre_port : herre_config["port"]= herre_port
        if client_id : herre_config["client_id"]= client_id
        if client_secret : herre_config["client_secret"]= client_secret
        if grant_type: herre_config["grant_type"] = grant_type

        
        if allow_insecure: 
            herre_config["secure"] = False
            arkitekt_config["secure"] = False
                
        herre_config = HerreConfig(**herre_config)
        arkitekt_config = ArkitektConfig(**arkitekt_config)


        if herre_config.authorization_grant_type == GrantType.CLIENT_CREDENTIALS: auth = ArnheimBackendOauth(herre_config, force_new_token=force_new_token)
        elif herre_config.authorization_grant_type == GrantType.IMPLICIT: auth = ImplicitApplication(herre_config, force_new_token=force_new_token)
        elif herre_config.authorization_grant_type == GrantType.PASSWORD: auth = LegacyApplication(herre_config, username=username, password=password, force_new_token=force_new_token)
        else: raise NotImplementedError("Please Specifiy a valid Grant Type")

        super().__init__(auth, arkitekt_config, auto_negotiate=True, bind=bind, **kwargs)


from typing import List
from PyQt5.QtWidgets import  QDialog, QDialogButtonBox, QFormLayout, QGroupBox,  QVBoxLayout 
from bergen.schema import Port
from functools import partial

class AssignationUI(QDialog):


    def __init__(self, inputs: List[Port], default_values = {}, forced_values={}, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        
        
        self.inputs = inputs
        self.inputs_with_widgets = [port for port in inputs if port.widget is not None]



        self.editable_ports = [port for port in self.inputs_with_widgets if port.key not in forced_values.keys()]

        self.keyDependencyMap = {port.key: port.widget.dependencies for port in self.inputs_with_widgets}

        self.keyWidgetMap = {
            port.key: port.buildWidget(on_changed=partial(self.onDependencyChanged, port.key), parent=self) for port in self.editable_ports
        }

        self.keyLabelMap = {
            port.key: port.label for port in self.editable_ports
        }

        self.defaultValuesMap = {
            port.key: port.default for port in self.inputs
        }

        self.keyValuesMap = {**self.defaultValuesMap, **default_values, **forced_values}


        #UI

        self.parametersWidget = QGroupBox("Parameters")
        self.parametersLayout = QFormLayout()

        self.port_widgets = [widget for widget in self.keyWidgetMap.values()]
        self.port_labels = [label for label in self.keyLabelMap.values()]

        for row in zip(self.port_labels, self.port_widgets):
            self.parametersLayout.addRow(*row)

        self.parametersWidget.setLayout(self.parametersLayout)

        buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        buttonBox.accepted.connect(self.on_set)
        buttonBox.rejected.connect(self.reject)

        self.layout = QVBoxLayout()
        self.layout.addWidget(self.parametersWidget)
        self.layout.addWidget(buttonBox)
        self.setLayout(self.layout)

    
        self.initialRender()


    def on_set(self):
        self.parameters = self.keyValuesMap
        self.accept()
        self.close()


    def initialRender(self):

        no_dependency_keys = [key for key, dependencies in self.keyDependencyMap.items() if len(dependencies) == 0]

        for key in no_dependency_keys:
            if key in self.keyWidgetMap:
                self.keyWidgetMap[key].render(self.keyValuesMap)
    

    def onDependencyChanged(self, updatedkey, value):
        if self.keyValuesMap[updatedkey] == value:
            # No change in dependency, omit
            return
        else:
            self.keyValuesMap[updatedkey] = value
            dependent_keys = [key for key, dependencies in self.keyDependencyMap.items() if updatedkey in dependencies]
            print(dependent_keys)
            
            for key in dependent_keys:
                self.keyWidgetMap[key].render(self.keyValuesMap)





from abc import abstractmethod
from abc import ABC
import asyncio
from bergen.schema import DataPoint
from bergen.query import  TypedGQL
from typing import TypeVar
from bergen.console import console


T = TypeVar("T")

class WardException(Exception):
    pass

class ConnectionError(Exception):
    pass




class BaseWard(ABC):

    def __init__(self, client, point: DataPoint, loop=None):
        self.loop = loop or client.loop or asyncio.get_event_loop()

        self.needs_negotiation = point.needsNegotiation
        self.host = point.outward or client.config.host
        self.port = point.port or client.config.port
        self.protocol = "https" if point.secure or client.config.secure else "http"

        self.token = client.auth.access_token
        assert self.token is not None, "Cannot create a Ward without acquiring a Token first"


    @abstractmethod
    async def connect(self):
        pass


    async def configure(self):
        try:
            await self.connect()
            if self.needs_negotiation: 
                return await self.negotiate()
        except:
            console.print_exception()
            raise ConnectionError(f"Connection to {self.host}:{self.port} on {self.port} Failed")
        


    def run(self, the_query: TypedGQL, variables: dict = {}, **kwargs):
        return self.loop.run_until_complete(self.run_async(the_query, variables=variables, **kwargs))


    @abstractmethod
    def run_async(self, gql: TypedGQL, variables: dict = {}):
        return gql.cls(**{})


    @abstractmethod
    async def disconnect(self):
        pass


    async def __aenter__(self):
        await self.configure()
        return self


    async def __aexit__(self, *args, **kwargs):
        await self.disconnect()





from bergen.schemas.herre.types import Application, User
from bergen.config.types import HerreConfig
import json
import logging
import os
import shelve
from abc import ABC, abstractmethod
import requests


logger = logging.getLogger(__name__)

from bergen.console import console


class AuthError(Exception):
    pass



class BaseAuthBackend(ABC):


    def __init__(self, config: HerreConfig, token_url="o/token/", authorize_url="o/authorize/", check_endpoint="auth/", force_new_token=False) -> None:
        # Needs to be set for oauthlib
        os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "0" if config.secure else "1"
        if not config.secure: console.log("Using Insecure Oauth2 Protocol.. Please only for local and debug deployments")


        self.config = config
        self.base_url = f'{"https" if config.secure else "http"}://{config.host}:{config.port}/'
        self.check_url = self.base_url + check_endpoint
        self.auth_url = self.base_url + authorize_url
        self.token_url = self.base_url + token_url
        self.scopes = config.scopes + ["introspection"]
        self.client_id = config.client_id
        self.client_secret = config.client_secret  
        self.force_new_token = force_new_token  

        self.scope = " ".join(self.scopes)
        
        
        self._user = None
        self._accesstoken = None
        self._application = None

        # Lets check if we already have a local toke
        config_name = "token.db"
        run_path = os.path.abspath(os.getcwd())
        self.db_path = os.path.join(run_path, config_name)
        

        self.token = None
        self.needs_validation = False

        if not self.force_new_token:
            try:
                with shelve.open(self.db_path) as cfg:
                        self.token = cfg['token']
                        self.needs_validation = True
                        logger.debug("Found local config")
            except KeyError:
                logger.info("No configuration found")

        super().__init__()


    @abstractmethod
    def fetchToken(self, loop=None) -> str:
        raise NotImplementedError("This is an abstract Class")

    def refetchToken(self) -> str:
        raise NotImplementedError("This is an abstract Class")

    def getUser(self):
        assert self.token is not None, "Need to authenticate before accessing the User"
        if not self._user:
            answer = requests.get(self.base_url + "me/", headers={"Authorization": f"Bearer {self.access_token}"})
            self._user = User(**answer.json())
        return self._user

    @property
    def access_token(self) -> str:
        if not self._accesstoken:
            self.getToken()
        
        return self._accesstoken

    @property
    def user(self) -> User:
        if not self._user:
            self.getUser()

        return self._user

    @property
    def application(self) -> Application:
        if not self._application:
            self.getToken()
            
        return self._application


    def getToken(self, loop=None) -> str:
        if self._accesstoken is None:
            if self.token is None:
                try:
                    self.token = self.fetchToken()
                except ConnectionResetError:
                    raise Exception(f"We couldn't connect to the Herre instance at {self.base_url}")     
                except:
                    logger.error(f"Couldn't fetch Token with config {self.config}")
                    raise
                
                with shelve.open(self.db_path) as cfg:
                    cfg['token'] = self.token


            response = requests.get(self.check_url, headers={"Authorization": f"Bearer {self.token['access_token']}"})

            if response.status_code != 200:
                self.token = self.fetchToken()
                response = requests.get(self.check_url, headers={"Authorization": f"Bearer {self.token['access_token']}"})


            assert response.status_code == 200, "Was not able to get a valid token"
            self._application = Application(**json.loads(response.content))
            self._accesstoken =  self.token["access_token"]

        return self._accesstoken



import asyncio
from typing import Any
from bergen.contracts.reservation import Reservation
from bergen.contracts.interaction import Interaction

from bergen.monitor.monitor import Monitor
from bergen.messages.postman.reserve.bounced_reserve import ReserveParams
from bergen.schema import AssignationParams, Node
from bergen.registries.client import get_current_client
from bergen.contracts import Reservation
from aiostream import stream
from tqdm import tqdm
import textwrap
import logging
from rich.table import Table
from rich.table import Table


logger = logging.getLogger(__name__)

class AssignationUIMixin:

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._ui = None


    def askInputs(self, **kwargs) -> dict:
        widget = self.getWidget(**kwargs) # We have established a ui
        if widget.exec_():
            return widget.parameters
        else:
            return None


    def getWidget(self, **kwargs):
        try:
            from bergen.ui.assignation import AssignationUI
            if not self._ui:
                self._ui = AssignationUI(self.inputs, **kwargs)
            return self._ui
        except ImportError as e:
            raise NotImplementedError("Please install PyQt5 in order to use interactive Widget based parameter query")



class NodeExtender(AssignationUIMixin):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args,**kwargs)
        
        bergen = get_current_client()

        self._postman = bergen.getPostman()
        self._loop = bergen.loop


    def interactive(self) -> Interaction:
        return Interaction(self)

    def reserve(self, loop=None, monitor: Monitor = None, ignore_node_exceptions=False, bounced=None, **params) -> Reservation:
        return Reservation(self, loop=loop, monitor=monitor, ignore_node_exceptions=ignore_node_exceptions, bounced=bounced, **params)

    async def stream(self, inputs: dict, params: ReserveParams = None, **kwargs):
        return stream.iterate(self._postman.stream(self, inputs, params, **kwargs))

    def _repr_html_(self: Node):
        string = f"{self.name}</br>"

        for arg in self.args:
            string += "Args </br>"
            string += f"Port: {arg._repr_html_()} </br>"

        for kwarg in self.kwargs:
            string += "Kwargs </br>"
            string += f"Port: {kwarg._repr_html_()} </br>"


        return string


    def __call__(self, *args: Any, **kwds: Any) -> Any:
        return super().__call__(*args, **kwds)


    def __rich__(self):
        my_table = Table(title=f"Node: {self.name}", show_header=False)

        my_table.add_row("ID", str(self.id))
        my_table.add_row("Package", self.package)
        my_table.add_row("Interface", self.interface)

        return my_table


from bergen.models import Node






def use(package=None, interface=None) -> Node:
    """Use a Node on the Platform by Searching for it on its package

    Args:
        package ([type], optional): The package this Node belongs to. Defaults to None.
        interface ([type], optional):  The interface of this Node. Defaults to None.

    Returns:
        Node: The Node
        
    """
    return Node.objects.get(package=package, interface=interface)






import asyncio
from asyncio.tasks import Task
from bergen.debugging import DebugLevel
from bergen.handlers.base import Connector
from bergen.messages import *
from bergen.handlers import *
from bergen.console import console
from bergen.utils import *


class Actor:
    expandInputs = True
    shrinkOutputs = True

    def __init__(self, connector: Connector, queue:asyncio.Queue = None, loop=None) -> None:
        self.queue = queue or asyncio.Queue()
        self.connector = connector
        self.loop = loop or asyncio.get_event_loop()
        self._provided = False

        self.assignments = {}
        self.reservations = {} # maps the reservations with its provided context

        self.reserve_handler_map = {}
        self.unreserve_handler_map = {}

        self.assign_handler_map = {}

        self.template = None
        pass

    async def log(self, message, level="INFO"):
        console.log(f"[red] Actor: {message}")



    async def on_provide(self, provide: ProvideHandler):
        pass

    async def on_unprovide(self, message):
        pass


    async def on_reserve(self, handler: ReserveHandler) -> Any:
        return None

    async def on_unreserve(self, unreserve_handler: UnreserveHandler, reserve_handler: ReserveHandler) -> None:
        return None

    
    async def _on_provide_message(self, bounced_provide: BouncedProvideMessage):
        self.provide_handler = ProvideHandler(bounced_provide, self.connector)
        try:
            self._current_provision_context = await self.on_provide(self.provide_handler)
            self.template = await self.provide_handler.get_template()
            self._provided = True
        except Exception as e:
            await self.provide_handler.pass_exception(e)


    async def _on_reserve_message(self, message: BouncedForwardedReserveMessage):
        reserve_handler = ReserveHandler(message, self.connector)
        self.reserve_handler_map[message.meta.reference] = reserve_handler

        try:
            await reserve_handler.log("Reserving", level=DebugLevel.INFO)

            reservation_context = await self.on_reserve(reserve_handler)
            if reservation_context is not None:
                reserve_handler.set_context(reservation_context)

            await reserve_handler.log("Reserving Done", level=DebugLevel.INFO)
            await reserve_handler.pass_done()

        except Exception as e:
            console.print_exception()
            await reserve_handler.pass_exception(e)

    async def _on_unreserve_message(self, message: BouncedUnreserveMessage):
        unreserve_handler = UnreserveHandler(message, self.connector)
        self.unreserve_handler_map[message.meta.reference] = unreserve_handler

        try:
            await unreserve_handler.log("Unreserving", level=DebugLevel.INFO)
            info = await self.on_unreserve(unreserve_handler, self.reserve_handler_map[message.data.reservation])
            del self.reserve_handler_map[message.data.reservation]
            await unreserve_handler.pass_done()
        except Exception as e:
            console.print_exception()
            await unreserve_handler.pass_exception(e)

            
    async def run(self, bounced_provide: BouncedProvideMessage):
        ''' An infinitie loop assigning to itself'''
        try:

            await self._on_provide_message(bounced_provide)

            try:
                assert self._provided, "We didnt provide this actor before running"
                
                while True:
                    message = await self.queue.get()

                    if isinstance(message, BouncedForwardedReserveMessage):
                        await self._on_reserve_message(message)

                    elif isinstance(message, BouncedUnreserveMessage):
                        await self._on_unreserve_message(message)

                    elif isinstance(message, BouncedForwardedAssignMessage):
                        task = asyncio.create_task(self.on_assign(message))
                        task.add_done_callback(self.check_if_assignation_cancelled)
                        self.assignments[message.meta.reference] = task

                    elif isinstance(message, BouncedUnassignMessage):
                        if message.data.assignation in self.assignments: 
                            await self.log(f"Cancellation of assignment {message.data.assignation}", level=DebugLevel.INFO)
                            task = self.assignments[message.data.assignation]
                            if not task.done():
                                task.cancel()
                                unassign_done = UnassignDoneMessage(data={
                                    "assignation": message.data.assignation
                                },
                                    meta = {
                                        "reference": message.meta.reference,
                                        "extensions": message.meta.extensions
                                    }
                                )
                                await self.connector.forward(unassign_done)
                                await self.log(f"Cancellation of assignment suceeded", level=DebugLevel.INFO)
                            else:
                                unassign_critical = UnassignCriticalMessage(data={
                                    "message": "Task was already done",
                                    "type": "Exception"
                                },
                                    meta = {
                                        "reference": message.meta.reference,
                                        "extensions": message.meta.extensions
                                    }
                                )
                                await self.connector.forward(unassign_critical)
                                await self.log(f"Cancellation of assignment failed. Task was already Done", level=DebugLevel.INFO)
                                #TODO: Maybe send this to arkitekt as well?
                        else:
                            raise Exception("Assignment never was at this pod. something went wrong")

                    else:
                        raise Exception(f"Type not known {message}")

                    self.queue.task_done()


            except asyncio.CancelledError:
                await self.on_unprovide(None)
                await self.log("Actor was beeing cancelled")
                raise

        except Exception as e:
            console.print_exception()

    
    def check_if_assignation_cancelled(self, task: Task):
        if task.cancelled():
            console.log(f"[yellow] Assignation {task.get_name()} Cancelled and is now Done")
        elif task.exception():
            console.log(f"[red] Assignation {task.get_name()} Failed with {str(task.exception())}")
        elif task.done():
            console.log(f"[green] Assignation {task.get_name()} Succeeded and is now Done")


    async def on_assign(self, assign: BouncedForwardedAssignMessage):
        assign_handler = AssignHandler(message=assign, connection=self.connector)
        self.assign_handler_map[assign.meta.reference] = assign_handler
        reserve_handler = self.reserve_handler_map[assign.data.reservation]

        await assign_handler.log(f"Assignment received", level=DebugLevel.INFO)
        try:
            try:
                args, kwargs = await expandInputs(node=self.template.node, args=assign.data.args, kwargs=assign.data.kwargs) if self.expandInputs else (assign.data.args, assign.data.kwargs)
           
                await self._assign(assign_handler, reserve_handler, args, kwargs) # We dont do all of the handling here, as we really want to have a generic class for Generators and Normal Functions
            
            except Exception as e:
                # As broad as possible to send further
                console.print_exception()
                await assign_handler.pass_exception(e)
                # Pass this further up
                raise

        except asyncio.CancelledError as e:
            await assign_handler.log(f"Cancellation of Assignment suceeded", level=DebugLevel.INFO)
            await assign_handler.pass_cancelled()
            raise e
    
    async def __aenter__(self):
        await self.reserve()
        return self

    async def __aexit__(self):
        await self.unreserve()



from bergen.query import TypedGQL
from bergen.schema import Transcript


NEGOTIATION_GQL = TypedGQL("""

  mutation Negotiate(
      $clientType: ClientTypeInput!
      $inward: String,
      $outward: String,
      $port: Int,
      $version: String,
      $pointType: DataPointTypeInput,
      $needsNegotiation: Boolean
      
      ) {
  negotiate(
      clientType: $clientType,
      inward: $inward,
      outward: $outward,
      port: $port,
      version: $version,
      pointType: $pointType,
      needsNegotiation: $needsNegotiation
  ) {
    timestamp
    models {
        identifier
        point {
            app {
                name
            }
            id
            outward
            port
            type
            needsNegotiation
        }
    }
    postman {
        type
        kwargs
    }
    provider {
        type
        kwargs
    }
    host {
        type
        kwargs
    }
  }
  }
""", Transcript)


from bergen.messages.postman.assign.assign_cancelled import AssignCancelledMessage
from bergen.messages.postman.progress import ProgressLevel
from bergen.debugging import DebugLevel
from bergen.handlers.base import ContractHandler
from bergen.messages import BouncedForwardedAssignMessage, AssignYieldsMessage, AssignProgressMessage, AssignDoneMessage, AssignReturnMessage, AssignCriticalMessage
from bergen.console import console


class AssignHandler(ContractHandler[BouncedForwardedAssignMessage]):

    @property
    def meta(self):
        return {"extensions": self.message.meta.extensions, "reference": self.message.meta.reference}

    async def log(self, message, level =ProgressLevel.INFO):
        console.log(message)
        await self.pass_progress(str(message), level=level)

    async def pass_yield(self, value):
        yield_message = AssignYieldsMessage(data={"returns": value}, meta=self.meta)
        await self.forward(yield_message)

    async def pass_progress(self, message: str, level: ProgressLevel = ProgressLevel.INFO):
        progress_message = AssignProgressMessage(data={"message": message, "level": level.value}, meta=self.meta)
        await self.forward(progress_message)

    async def pass_done(self):
        return_message = AssignDoneMessage(data={"ok": True}, meta=self.meta)
        await self.forward(return_message)

    async def pass_cancelled(self):
        return_message = AssignCancelledMessage(data={"ok": True}, meta=self.meta)
        await self.forward(return_message)

    async def pass_result(self, value):
        return_message = AssignReturnMessage(data={"returns": value}, meta=self.meta)
        await self.forward(return_message)

    async def pass_exception(self, exception):
        console.print_exception()
        error_message = AssignCriticalMessage(data={"message": str(exception), "type": str(exception.__class__.__name__)}, meta=self.meta)
        await self.forward(error_message)


from abc import ABC, abstractmethod
from asyncio.futures import Future
from bergen.messages.postman.progress import ProgressLevel
from bergen.messages import *
from bergen.messages.base import MessageModel
from bergen.postmans.utils import build_assign_message, build_reserve_message, build_unassign_messsage, build_unreserve_messsage
import uuid
from bergen.hookable.base import Hookable
from bergen.schema import Node, Pod, Template
from typing import Callable, Dict, List
from aiostream import stream
import asyncio
import logging
from bergen.console import console

logger = logging.getLogger(__name__)

ReferenceQueueMap = Dict[str, asyncio.Queue]
ReferenceFutureMap = Dict[str, Future]
ReferenceProgressFuncMap = Dict[str, Callable]

class NodeException(Exception):
    pass

class HostException(Exception):
    pass

class ProtocolException(Exception):
    pass

class BasePostman(Hookable):
    """ A Postman takes node requests and translates them to Bergen calls, basic implementations are GRAPHQL and PIKA"""
    
    def __init__(self, client, requires_configuration=True, loop=None,**kwargs) -> None:
        super().__init__(**kwargs)
        self.loop = loop or client.loop or asyncio.get_event_loop()
        self.client = client

        # Assignments and their Cancellations
        self.assignment_stream_queues: ReferenceQueueMap = {}
        self.assignments: ReferenceFutureMap = {}
        self.assignment_progress_functions: ReferenceProgressFuncMap = {}

        self.unassignments: ReferenceFutureMap = {}
        self.unassignment_progress_functions: ReferenceProgressFuncMap = {}

        # Reservations and their Cancellations
        self.reservation_stream_queues: ReferenceQueueMap = {}
        self.reservations: ReferenceFutureMap = {}
        self.reservations_progress_functions: ReferenceProgressFuncMap = {}

        self.unreservations: ReferenceFutureMap = {}
        self.unreservations_progress_functions: ReferenceProgressFuncMap = {}



    @abstractmethod
    async def connect(self):
        pass

    @abstractmethod
    async def disconnect(self):
        pass

    @abstractmethod
    async def forward(self, message: MessageModel):
        return NotImplementedError("This is abstract")


    async def on_message(self, message: MessageModel):
        # First we check the streams
        reference = message.meta.reference

        if reference in self.assignment_stream_queues:
            return await self.assignment_stream_queues[reference].put(message)

        if reference in self.reservation_stream_queues:
            return await self.reservation_stream_queues[reference].put(message)


        # Assign Function related
        if isinstance(message, AssignReturnMessage):
            future = self.assignments.pop(reference)
            if not future.cancelled():
                future.set_result(message)
            else:
                if reference in self.assignment_progress_functions:
                    function = self.assignment_progress_functions[reference]
                    function(f"[red]Race Condition! Assignation Return before being Cancelled on the Server. Omitting Result!", level=ProgressLevel.DEBUG)

        elif isinstance(message, AssignProgressMessage):
            if reference in self.assignment_progress_functions:
                function = self.assignment_progress_functions[reference]
                function(message.data.message, level=message.data.level)
            else:
                logger.warning(f"Received unwanted Progress {message}")

        elif isinstance(message, AssignCriticalMessage):
            future = self.assignments.pop(reference)
            future.set_exception(NodeException(message.data.message))


        # Unassign Function related
        elif isinstance(message, UnassignDoneMessage):
            future = self.unassignments.pop(reference)
            future.set_result(message)

        elif isinstance(message, UnassignProgressMessage):
            if reference in self.unassignment_progress_functions:
                function = self.unassignment_progress_functions[reference]
                function(message.data.message, level=message.data.level)
            else:
                logger.warning(f"Received unwanted Progress {message}")

        elif isinstance(message, UnassignCriticalMessage):
            future = self.unassignments.pop(reference)
            future.set_exception(NodeException(message.data.message))


        # Reserve Function related
        elif isinstance(message, ReserveDoneMessage):
            future = self.reservations.pop(reference)
            future.set_result(message)

        elif isinstance(message, ReserveProgressMessage):
            if reference in self.reservations_progress_functions:
                function = self.reservations_progress_functions[reference]
                function(message.data.message, level=message.data.level)
            else:
                logger.warning(f"Received unwanted Progress {message}")

        elif isinstance(message, ReserveCriticalMessage):
            future = self.reservations.pop(reference)
            future.set_exception(HostException(message.data.message))


        # Reserve Function related
        elif isinstance(message, UnreserveDoneMessage):
            future = self.unreservations.pop(reference)
            future.set_result(message)

        elif isinstance(message, UnreserveProgressMessage):
            if reference in self.unreservations_progress_functions:
                function = self.unreservations_progress_functions[reference]
                function(message.data.message, level=message.data.level)
            else:
                logger.warning(f"Received unwanted Progress {message}")


        elif isinstance(message, UnreserveCriticalMessage):
            future = self.unreservations.pop(reference)
            future.set_exception(HostException(message.data.message))


        elif isinstance(message, ExceptionMessage):
            if reference in self.assignments: self.assignments.pop(reference).set_exception(message.toException()) 
            if reference in self.unassignments: self.unassignments.pop(reference).set_exception(message.toException()) 
            if reference in self.reservations: self.reservations.pop(reference).set_exception(message.toException()) 
            if reference in self.unreservations: self.unreservations.pop(reference).set_exception(message.toException())

        else:

            raise Exception("Unknown message type", message )
        


    async def unassign(self, assignation: str, on_progress=None, bounced=None) -> UnassignDoneMessage:
        """ Takes a previously assigned reference to an assignation and cancel the assignation

        Args:
            assignation (str): The reference to the assignation
            on_progress ([type], optional): A callacble for the progress. Defaults to None.

        Raises:
            e: A cancellation Error

        Returns:
            None
        """
        unassign_reference = str(uuid.uuid4()) 

        future = self.loop.create_future()
        self.unassignments[unassign_reference] = future


        with_progress = False
        if on_progress:
            assert callable(on_progress), "on_progress if provided must be a function/lambda"
            self.unassignment_progress_functions[unassign_reference] = on_progress
            with_progress = True

        unassign = build_unassign_messsage(unassign_reference,assignation, with_progress=with_progress, bounced=bounced)


        await self.forward(unassign)

        try:
            future.add_done_callback(lambda x: logger.info(x))
            return await future

        except asyncio.CancelledError as e:
            if on_progress: on_progress(f"[red]Cancelled Unassignment {unassign_reference}")
            raise e



    async def assign(self, reservation: str, shrinked_args, shrinked_kwargs = {}, on_progress: Callable = None, bounced = None) -> AssignReturnMessage:
        """Assign takes a reservation and the serialized arguments as well as kwargs and awaits the result call, it will return
        the Assignation Returns.

        Attention! This function should only be called within an ongoing reservation

        Args:
            reservation (str): The reference of the reservation you want to assign to
            serialized_args ([type]): The serialized Args
            serialized_kwargs (dict, optional): [description]. The serialized Kwargs Defaults to {}.
            on_progress (Callable, optional): A on_progress callable for Assign and Unassign Progress. Defaults to None.
            bounced (Dict, optional): A dict using the bounced context from another assignment (works only on backend applications with can_forward_bounce scope

        Raises:
            e: Cancellation Error

        Returns:
            List: The Unserialized Outputs of the Call
        """
        
        assign_reference = str(uuid.uuid4()) 

        future = self.loop.create_future()
        self.assignments[assign_reference] = future

        with_progress = False
        if on_progress:
            assert callable(on_progress), "on_progress if provided must be a function/lambda"
            self.assignment_progress_functions[assign_reference] = on_progress
            with_progress = True
            
        assign = build_assign_message(assign_reference, reservation, shrinked_args, kwargs=shrinked_kwargs, with_progress=with_progress, bounced=bounced)
        await self.forward(assign)

        try:
            future.add_done_callback(lambda x: logger.info(x))
            return await future


        except asyncio.CancelledError as e:
            if on_progress: on_progress(f"[red]Cancelled Assignment {assign_reference}", level=ProgressLevel.DEBUG)
            try:
                unassign = await self.unassign(assign_reference, bounced=bounced)
            except:
                console.print_exception()
            raise e


    async def unreserve(self, reservation, on_progress: Callable = None, bounced=None) -> UnreserveDoneMessage:
        unreserve_reference = str(uuid.uuid4()) 

        future = self.loop.create_future()
        self.unreservations[unreserve_reference] = future


        with_progress = False
        if on_progress:
            assert callable(on_progress), "on_progress if provided must be a function/lambda"
            self.unreservations_progress_functions[unreserve_reference] = on_progress
            with_progress = True

        unassign = build_unreserve_messsage(unreserve_reference, reservation, with_progress=with_progress, bounced=bounced)
        await self.forward(unassign)

        try:
            future.add_done_callback(lambda x: logger.info(x))
            return await future

        except asyncio.CancelledError as e:
            if on_progress: on_progress(f"[red]Cancelled Unreservation {unreserve_reference}", level=ProgressLevel.DEBUG)
            raise e



    async def reserve(self, node_id: str = None, template_id: str = None , params_dict: dict = {}, on_progress: Callable = None, bounced=None) -> ReserveDoneMessage:
        """[summary]

        Args:
            node_id (str, optional): [description]. Defaults to None.
            template_id (str, optional): [description]. Defaults to None.
            params_dict (dict, optional): [description]. Defaults to {}.
            on_progress (Callable, optional): [description]. Defaults to None.

        Raises:
            e: [description]

        Returns:
            str: The 
        """

        reserve_reference = str(uuid.uuid4()) 

        future = self.loop.create_future()
        self.reservations[reserve_reference] = future

        with_progress = False
        if on_progress:
            assert callable(on_progress), "on_progress if provided must be a function/lambda"
            self.reservations_progress_functions[reserve_reference] = on_progress
            with_progress = True
            
        reserve = build_reserve_message(reserve_reference, node_id, template_id, params_dict=params_dict, with_progress=with_progress,bounced=bounced)
        await self.forward(reserve)

        try:
            future.add_done_callback(lambda x: logger.info(x))
            reserve_done: ReserveDoneMessage = await future
            await reserve_done

        except asyncio.CancelledError as e:
            if on_progress: on_progress(f"[red]Cancelled Reservation {reserve_reference}")
            unassign = await self.unreserve(self, reserve_reference, bounced=bounced)
            raise e

    async def reserve_stream(self, node_id: str = None, template_id: str = None , params_dict: dict = {}, with_progress= True, bounced=None) -> MessageModel:
        reserve_reference = str(uuid.uuid4())
        self.reservation_stream_queues[reserve_reference] = asyncio.Queue()


        reserve = build_reserve_message(reserve_reference, node_id, template_id, params_dict=params_dict, with_progress=with_progress, bounced=bounced)
        await self.forward(reserve)

        try:
            while True:
                parsed_message = await self.reservation_stream_queues[reserve_reference].get()
                yield parsed_message

        except asyncio.CancelledError as e:
            # Otherwise we will still listen to the stream on cancellation
            del self.reservation_stream_queues[reserve_reference]
            unreserve = await self.unreserve(reserve_reference, bounced=bounced)
            raise e # Otherwise we


    async def assign_stream(self, reservation: str, serialized_args, serialized_kwargs = {}, with_progress = None, bounced=None) -> MessageModel:
        assign_reference = str(uuid.uuid4())
        self.assignment_stream_queues[assign_reference] = asyncio.Queue()

        assign = build_assign_message(assign_reference, reservation, serialized_args, kwargs=serialized_kwargs, with_progress=with_progress, bounced=bounced)
        await self.forward(assign)

        try:
            while True:
                parsed_message = await self.assignment_stream_queues[assign_reference].get()
                yield parsed_message
                
        except asyncio.CancelledError as e:
            # Otherwise we will still listen to the stream on cancellation
            del self.assignment_stream_queues[assign_reference]
            unassign = await self.unassign(self, assign_reference, bounced=bounced)
            # TODO: Maybe yield here as well?
            raise e # Otherwise we